@@ -11,6 +11,7 @@ import threading
 
 import sqlalchemy
 from sqlalchemy.engine.url import make_url
+from sqlalchemy.orm import sessionmaker
 
 from zope.interface import implements
 from zope.component import getUtility
@@ -21,7 +22,7 @@ from z3c.sqlalchemy.model import Model
 from z3c.sqlalchemy.mapper import LazyMapperCollection
 
 import transaction
-from transaction.interfaces import IDataManager
+from transaction.interfaces import ISavepointDataManager, IDataManagerSavepoint
 
 
 class SynchronizedThreadCache(object):
@@ -30,31 +31,37 @@ class SynchronizedThreadCache(object):
         self.lock = threading.Lock()
         self.cache = threading.local()
 
-
-    def set(self, **kw):
+    def set(self, id, d):
         self.lock.acquire()
-        for k,v in kw.items():
-            setattr(self.cache, k, v)
+        setattr(self.cache, id, d)
         self.lock.release()
 
-
-    def get(self, *names):
+    def get(self, id):
         self.lock.acquire()
-        result = [getattr(self.cache, name, None) for name in names]
+        result = getattr(self.cache, id, None)
         self.lock.release()
         return result
 
+    def remove(self, id):
+        self.lock.acquire()
+        if hasattr(self.cache, id):
+            delattr(self.cache, id)           
+        self.lock.release()
+
+
 
 class BaseWrapper(object):
 
     implements(ISQLAlchemyWrapper)
 
-    def __init__(self, dsn, model=None, transactional=True, **kw):
+    def __init__(self, dsn, model=None, transactional=True, engine_options={}, session_options={}, **kw):
         """ 'dsn' - a RFC-1738-style connection string
 
             'model' - optional instance of model.Model
 
-            'kw' - optional keyword arguments passed to create_engine()
+            'engine_options' - optional keyword arguments passed to create_engine()
+
+            'session_options' - optional keyword arguments passed to create_session() or sessionmaker()
 
             'transactional' - True|False, only used by SQLAlchemyDA, 
                               *don't touch it*
@@ -69,11 +76,13 @@ class BaseWrapper(object):
         self.dbname = self.url.database 
         self.drivername = self.url.drivername
         self.transactional = transactional
-        self.kw = kw
-        self.echo = kw.get('echo', False)
+        self.engine_options = engine_options
+        if 'echo' in kw:
+            self.engine_options.update(echo=kw['echo'])
+        self.session_options = session_options
         self._model = None
         self._createEngine()
-        self._id = random.random() # used as unique key for session/connection cache
+        self._id = str(random.random()) # used as unique key for session/connection cache
 
         if model:
 
@@ -113,7 +122,7 @@ class BaseWrapper(object):
 
     @property
     def session(self):
-        return sqlalchemy.create_session(self._engine)
+        return self._sessionmaker()
 
     def registerMapper(self, mapper, name):
         self._mappers.registerMapper(mapper, name)
@@ -135,38 +144,52 @@ class BaseWrapper(object):
         return self._model
 
     def _createEngine(self):
-        self._engine = sqlalchemy.create_engine(self.dsn, **self.kw)
-        self._engine.echo = self.echo
+        self._engine = sqlalchemy.create_engine(self.dsn, **self.engine_options)
+        self._sessionmaker = sqlalchemy.orm.sessionmaker(bind=self._engine, 
+                                                         autoflush=True,
+                                                         transactional=True,
+                                                         **self.session_options)
 
 
-session_cache = SynchronizedThreadCache()
 connection_cache = SynchronizedThreadCache()
 
 
 class SessionDataManager(object):
     """ Wraps session into transaction context of Zope """
 
-    implements(IDataManager)
+    implements(ISavepointDataManager)
 
-    def __init__(self, session, id):
+    def __init__(self, connection, session, id, transactional=True):
+
+        self.connection = connection
         self.session = session
+        self.transactional = True
         self._id = id
-        self.transaction = self.session.create_transaction()
+        self.transaction = None
+        if self.transactional:
+            self.transaction = connection.begin()
 
     def abort(self, trans):
-        if self.transaction is not None:
-            self.transaction.rollback()
-        self.session.clear()
-        session_cache.set(**{'last_session_%s' % self._id : None})
+
+        try:
+            if self.transaction is not None:
+                self.transaction.rollback()
+        # DM: done in "_cleanup" (similar untidy code at other places as well)
+##        self.session.clear()
+##        connection_cache.remove(self._id)
+        finally:
+            # ensure '_cleanup' is called even when 'rollback' causes an exception
+            self._cleanup()
 
     def _flush(self):
+
         # check if the session contains something flushable
         if self.session.new or self.session.deleted or self.session.dirty:
 
             # Check if a session-bound transaction has been created so far.
             # If not, create a new transaction
-            if self.transaction is None:
-                self.transaction = self.session.create_transaction()
+#            if self.transaction is None:
+#                self.transaction = connection.begin()
 
             # Flush
             self.session.flush()
@@ -181,60 +204,51 @@ class SessionDataManager(object):
         self._flush()
 
     def tpc_finish(self, trans):
+
         if self.transaction is not None:
             self.transaction.commit()
-        self.session.clear()
-        session_cache.set(**{'last_session_%s' % self._id : None})
 
-
-    def tpc_abort(self, trans):
-        if self.transaction is not None:
-            self.transaction.rollback()
         self.session.clear()
-        session_cache.set(**{'last_session_%s' % self._id : None})
+        self._cleanup()
+        
+
+    # DM: no need to duplicate this code (identical to "abort")
+##    def tpc_abort(self, trans):
+##        if self.transaction is not None:
+##            self.transaction.rollback()
+##        self._cleanup()
+    tpc_abort = abort
 
     def sortKey(self):
-        return 'z3c.sqlalchemy' + str(id(self))
+        return 'z3c.sqlalchemy_' + str(id(self))
+
+    def _cleanup(self):
+        self.session.clear()
+        if self.connection:
+            self.connection.close()
+            self.connection = None
+        connection_cache.remove(self._id)
+        # DM: maybe, we should set "transaction" to "None"?
+
+    def savepoint(self):
+        """ return a dummy savepoint """
+        return AlchemySavepoint()
 
 
-class ConnectionDataManager(object):
-    """ Wraps connection into transaction context of Zope """
 
-    implements(IDataManager)
+# taken from z3c.zalchemy
 
-    def __init__(self, connection, transactional=True):
-        self.connection = connection
-        self.transactional = transactional
-        self.transaction = connection.begin()
+class AlchemySavepoint(object):
+    """A dummy saveoint """
 
-    def abort(self, trans):
-        if self.transactional:
-            self.transaction.rollback()
-        self.connection.close()
-        self.connection = None
-        connection_cache.set(last_connection=None)
+    implements(IDataManagerSavepoint)
 
-    def commit(self, trans):
-        if self.transactional:
-            self.transaction.commit()
-        self.connection.close()
-        self.connection = None
-        connection_cache.set(last_connection=None)
-
-    def tpc_begin(self, trans):
+    def __init__(self):
         pass
 
-    def tpc_vote(self, trans):
+    def rollback(self):
         pass
 
-    def tpc_finish(self, trans):
-        pass
-
-    def tpc_abort(self, trans):
-        pass
-
-    def sortKey(self):
-        return 'z3c.sqlalchemy' + str(id(self))
 
 
 class ZopeBaseWrapper(BaseWrapper):
@@ -242,47 +256,40 @@ class ZopeBaseWrapper(BaseWrapper):
         the session with the transaction management of Zope.
     """
 
-    @property
-    def session(self):
 
-        last_session, = session_cache.get('last_session_%s' % self._id)
+    def __getOrCreateConnectionCacheItem(self, cache_id):
+
+        cache_item = connection_cache.get(cache_id)
 
         # return cached session if we are within the same transaction
         # and same thread
-        if last_session is not None:
-            return last_session
+        if cache_item is not None:
+            return cache_item
 
         # no cached session, let's create a new one
-        session = sqlalchemy.create_session(self._engine)
+        connection = self.engine.connect()
+        session = sessionmaker(connection)()
                                           
         # register a DataManager with the current transaction
-        transaction.get().join(SessionDataManager(session, self._id))
+        transaction.get().join(SessionDataManager(connection, session, self._id))
 
         # update thread-local cache
-        session_cache.set(**{'last_session_%s' % self._id : session})
+        cache_item = dict(connection=connection, session=session)
+        connection_cache.set(self._id, cache_item)
+        return cache_item
+
+
+    @property
+    def session(self):
+        """ Return a (cached) session object for the current transaction """
+        return self.__getOrCreateConnectionCacheItem(self._id)['session']
 
-        # return the session
-        return session 
 
     @property
     def connection(self):
-
-        last_connection, = connection_cache.get('last_connection')
-
-        # return cached connection if we are within the same transaction
-        # and same thread
-        if last_connection is not None:
-            return last_connection
-
-        # no cached connection, let's create a new one
-        connection = self.engine.connect()
-                                          
-        # register a DataManager with the current transaction
-        transaction.get().join(ConnectionDataManager(connection, self.transactional))
-
-        # update thread-local cache
-        connection_cache.set(last_connection=connection)
-
-        # return the connection
-        return connection
-
+        """ This property is _private_ and only intented to be used
+            by SQLAlchemyDA and therefore it is not part of the 
+            public API. 
+        """
+    
+        return self.__getOrCreateConnectionCacheItem(self._id)['connection']
